import argparse
import os
from pathlib import Path

from venvlink import VenvLink
from venvlink.config import init_config
from venvlink.__version__ import __version__

parser = argparse.ArgumentParser(prog="venvlink", description=f"venvlink {__version__}")

parser.add_argument(
    "--init",
    help="Initiate the venvlink configuration file (.venvlinkrc)",
    action="store_true",
    default=False,
)

parser.add_argument(
    "-d",
    "--delete",
    help="Delete the virtual environment associated with project_name (instead of creating)",
    action="store_true",
    default=False,
)
parser.add_argument(
    "-S",
    "--system-site-packages",
    help="Give the virtual environment access to the system site-packages dir.",
    action="store_true",
    default=False,
)
parser.add_argument("projectname", action="store", nargs="?")


def create_venv(args):
    if args.projectname is None:
        print(
            f'The positional argument "projectname" is required for creating a virtual environment!'
        )
        return
    print(f'Creating venv for "{args.projectname.strip()}"')
    vlink = VenvLink()
    vlink.create_venv(
        args.projectname.strip(),
        workdir=Path(os.getcwd()),
        system_site_packages=args.system_site_packages,
    )


def delete_env(args):
    if args.projectname is None:
        print(
            f'The positional argument "projectname" is required for deleting a virtual environment!'
        )
    else:
        print(f'Deleting venv for "{args.projectname.strip()}"')
        vlink = VenvLink()
        vlink.delete_env(args.projectname.strip())


# Executed with "python -m venvlink [params]"
if __name__ == "__main__":
    args = parser.parse_args()

    if args.init:
        init_config()
    elif args.delete:
        delete_env(args)
    else:
        create_venv(args)


from configparser import ConfigParser
from functools import partial
import io
import logging
import os
from pathlib import Path

from venvlink.utils import get_user_folder, is_windows, is_in_accepted_values, get_input
from venvlink.exceptions import ImproperlyConfiguredError


def get_venvlinkrc():
    """
    Returns
    -------
    venvlinkrc: pathlib.Path
        The path to .venvlinkrc file.
        (the file might not exist)
    """
    filename = ".venvlinkrc"
    venvlinkrc = get_user_folder() / filename
    return venvlinkrc


def is_valid_path_or_empty(x):
    x = x.strip()
    if not x:  # empty string
        return True
    try:
        path = Path(x)
    except Exception:
        return False
    if not path.is_absolute():
        return False
    return True


def get_default_config(interactive=False):
    default_venv_folder = str(get_user_folder() / "venvs")

    text = "Folder for saving virtual envs?"
    text += f" [Default: {default_venv_folder}]" + os.linesep

    value = get_input(text, is_valid_path_or_empty).strip()
    if value:
        default_venv_folder = value

    default_config = """
    [general]
    VENV_FOLDER = {DEFAULT_VENV_FOLDER}
    """.strip().format(
        DEFAULT_VENV_FOLDER=default_venv_folder
    )
    return default_config


def init_config():
    file = get_venvlinkrc()
    if file.exists():

        text = f"WARNING: The configuration file '{file}' for venvlink exists already. "
        text += f"If you continue, the existing file will be OVERRIDDEN "
        text += "with the default configuration file " + os.linesep
        print(text)
        question = "Override current configuration file with a new one?"

        value = None
        while value not in {"Y", "N"}:
            print(question)
            prompt = "[Y] Yes [N] No [S] Show current .venlinkrc\n"
            validator = partial(is_in_accepted_values, accepted_values={"Y", "N", "S"})
            value = get_input(prompt, validator).upper()
            if value == "S":
                print(f"\nCurrent configuration ({file}):")
                print("-" * 28)
                with open(file) as f:
                    print(f.read())

        if value == "N":
            print("Canceled.")
            return

    config = create_default_config()

    save_config_to_file(config, file)
    print("Saved default config to file:", file)


def create_default_config():
    config = ConfigParser()
    buf = io.StringIO(get_default_config())
    config.read_file(buf)
    return config


def save_config_to_file(config, file):
    with open(file, "w", encoding="utf-8") as f:
        config.write(f)


class Configuration:
    def __init__(self, file=None):
        """
        Parameters
        ----------
        file: None or pathlib.Path
            The path to the ".venvlinkrc" (the
            configuration file).
        """
        self.file = get_venvlinkrc() if file is None else file

        self.config = self.load_config()

    def load_config(self):
        config = ConfigParser()
        try:
            with open(self.file, encoding="utf-8") as f:
                config.read_file(f)

        except FileNotFoundError:
            logging.debug(
                "Configuration file for venvlink does not exist. Using default configuration, instead"
            )
            config = create_default_config()
            self.save_config(config)
        return config

    def save_config(self, config):
        logging.info("Saving %s", self.file)
        save_config_to_file(config, self.file)

    @property
    def venv_folder(self):
        return self.get_key(section="general", key="venv_folder")

    def get_key(self, section, key, strip_quotes=True):
        """
        Gets a key from a [section] from the venvlink
        configuration file.

        Parameters
        ----------
        section: str
            Name of the section.
        key: str
            Name of the key inside [section]
        strip_quotes: bool
            If True, drops single and double quotes from
            the both ends of the string.
        """
        try:
            val = self.config[section][key]
        except KeyError:
            raise ImproperlyConfiguredError(
                self._get_improperly_configured_error_text(section, key)
            )

        if strip_quotes:
            val = val.strip('"').strip("'")

        return val

    def _get_improperly_configured_error_text(self, section, key):
        text = f'The section "{section}" of {self.file} did not contain key "{key}"'
        text += os.linesep * 2
        text += f"To fix this, ensure that {self.file} contains" + os.linesep * 2
        text += f"[{section}]" + os.linesep
        text += f"{key} = <some_value>" + os.linesep
        return text



import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    datefmt="%m-%d %H:%M",
)

from functools import partial
import os
from pathlib import Path
import subprocess
import sys
import shutil

from venvlink.__version__ import __version__
from venvlink.config import Configuration
from venvlink.exceptions import UserAborted
from venvlink.utils import is_in_accepted_values, get_input

logger = logging.getLogger(__name__)


def get_scripts_dir(venv_dir):
    """
    Parameters
    ---------
    venv_dir: pathlib.Path
        The virtual environment root dir.
    """
    return venv_dir / "Scripts"


def get_from_scripts_dir(venv_dir, file="activate"):
    return get_scripts_dir(venv_dir) / file


def get_from_venv_dir(venv_dir, file="pyvenv.cfg"):
    return venv_dir / file


def get_venvlink_text(venv_project):
    text = f"The virtual environment is located at {venv_project}" + os.linesep
    text += f"Created with venvlink v.{__version__}"
    return text


class VenvLink:
    def __init__(self, config_file=None):
        config_file = Path(config_file) if config_file is not None else None
        self.config = Configuration(config_file)

    def delete_env(self, project_name):
        """
        Removes virtual environment created with create_env.

        Parameters
        ----------
        project_name: str
            The project name for which the virtual enviroment
            was created. This should be the virtual environment
            name in the folder where all the virtual environments
            are located.
        """
        folder = self.venv_folder / project_name
        if folder.exists():
            logger.info("Removing %s", folder)

            shutil.rmtree(folder)
        else:
            logger.info("Could not remove %s! No such folder.", folder)

    def create_venv(self, project_name, workdir, system_site_packages=False):
        """
        Parameters
        ----------
        project_name: str
            The project name for which the virtual enviroment
            is created. This will be the virtual environment
            name in the folder where all the virtual environments
            are located.
        workdir: pathlib.Path
            The working directory; the directory into
            which the "linked virtual environment" is created
        system_site_packages: bool
            The --system_site_packages of "python -m venv"
        """

        try:
            self._check_no_venv_in_workdir(workdir, project_name)
        except UserAborted:
            print("Canceled.")
            return

        try:
            # Create the virtual environment in the "centralized location"
            # Note: project_name may change
            project_name = self._create_venv_to_venv_folder(
                project_name, system_site_packages=system_site_packages
            )
        except UserAborted:
            print("Canceled.")
            return

        # Create the "proxied" virtual environment in the workdir
        self._create_venv_to_workdir(workdir, project_name)

    def _create_venv_to_venv_folder(self, project_name, system_site_packages=False):

        # Create the folder for all the virtual environments
        self.venv_folder.mkdir(exist_ok=True, parents=True)

        ret = self._check_that_venv_does_not_exist(project_name)

        if ret == "skipcreate":
            return project_name
        if ret == "newname":
            project_name = self._get_new_venv_name()

        subprocess_cmd = [sys.executable, "-m", "venv", project_name]
        if system_site_packages:
            subprocess_cmd.append("--system-site-packages")
        logger.info(
            'Running: "%s" with cwd="%s"', " ".join(subprocess_cmd), self.venv_folder
        )
        subprocess.run(subprocess_cmd, cwd=self.venv_folder)
        return project_name

    def __venv_exists(self, project_name):
        return (self.venv_folder / project_name).exists()

    def _check_that_venv_does_not_exist(self, project_name):
        """
        Returns
        ret: str
            'continue' 'skipcreate' or 'newname'
        """
        if not self.__venv_exists(project_name):
            return "continue"

        return self._when_venv_exists_already(project_name)

    def _get_new_venv_name(self):

        prompt = "Give a new projectname: "
        errortxt = "Sorry, that projectname exists already. "
        return get_input(
            prompt,
            validator=lambda x: not self.__venv_exists(x),
            on_validationerror=errortxt,
        )

    def _when_venv_exists_already(self, project_name):
        """
        When venv already exists with the project_name, ask
        user opinion.

        Returns
        -------
        ret: str
            'skipcreate' or 'newname'

        Raises
        ------
        UserAborted, if user aborted.

        """
        text = f'The virtual environment for a projectname "{project_name}" exists already. '
        text += f'If you use the name "{project_name}", you will SHARE the virtual environment '
        text += "with the other project(s) using the same name." + os.linesep
        text += "Continue?" + os.linesep
        print(text)

        prompt = "[Y] Yes [N] No, give new name. [A] Abort: "
        validator = partial(is_in_accepted_values, accepted_values={"Y", "N", "A"})
        value = get_input(prompt, validator).upper()

        if value == "Y":
            return "skipcreate"
        elif value == "N":
            return "newname"
        elif value == "A":
            raise UserAborted()

    def _check_no_venv_in_workdir(self, workdir, project_name, venvname="venv"):

        venvdir_dst = workdir / venvname
        if venvdir_dst.exists():
            print(f'The activate proxy inside "{venvdir_dst}" exists already!')
            print(
                "Do you want to remove it? (This will remove just the link, not the virtual environment it is linked to)"
            )

            prompt = "[Y] Yes [A] Abort: "
            validator = partial(is_in_accepted_values, accepted_values={"Y", "A"})
            value = get_input(prompt, validator).upper()

            if value == "A":
                raise UserAborted()
            elif value == "Y":
                print("Removing ", str(venvdir_dst))
                shutil.rmtree(venvdir_dst)
            else:  # should not ever happen.
                ValueError()

        return True

    def _create_venv_to_workdir(self, workdir, project_name, venvname="venv"):
        # Create the working directory if it does not
        # exists (should probably always exist, but anyway.)
        workdir.mkdir(exist_ok=True, parents=True)

        venvdir_dst = workdir / venvname
        venvdir_dst.mkdir()

        # Make proxies to needed files
        venvdir_src = self.venv_folder / project_name

        real_activate_ps1 = venvdir_src / "Scripts" / "Activate.ps1"
        real_activate_bat = venvdir_src / "Scripts" / "activate.bat"

        files_and_contents = (
            (
                "Scripts" + os.path.sep + "Activate.ps1",
                (
                    f"""Write-Output 'venvlink: Activating virtual env in "{venvdir_src}"'\n"""
                    f"& '{real_activate_ps1}'"
                ),
            ),
            (
                "Scripts" + os.path.sep + "activate.bat",
                (
                    "@echo off\n"
                    f"""echo venvlink: Activating virtual env in \"{venvdir_src}\"\n"""
                    f"""call \"{real_activate_bat}\""""
                ),
            ),  # Dummy file which makes possible to run simply "venv\Scripts\activate"
            ("Scripts" + os.path.sep + "activate", ""),
        )

        for file, content in files_and_contents:
            file_dst = venvdir_dst / file
            file_dst.parent.mkdir(exist_ok=True, parents=True)
            with open(file_dst, "w", encoding="utf-8") as f:
                print(content, file=f)

        with open(venvdir_dst / "venvlink", "w") as f:
            print(get_venvlink_text(venvdir_src), file=f)

    @property
    def venv_folder(self):
        return Path(self.config.venv_folder)



import os
from pathlib import Path 
import sys 
from venvlink.__version__ import __version__


def is_in_accepted_values(value, accepted_values):
    """
    Case-insensitive check for accepted values

    Parameters
    ---------
    value: str
        The string to be validated
    accepted_values: str
        The accepted values. Use a set of UPPERCASE
        strings.
    """
    if value.upper() in accepted_values:
        return True
    return False


def get_input(prompt, validator, on_validationerror=None):
    """
    Parameters
    ----------
    prompt: str
        The text to show to user when asking for input
    validator: function
        The function to call, with the user input as argument
        * function should return True, if the input is ok
        * function should return False, if input is not ok
    on_validationerror: str
        The text to show if validation with `func` fails
    """
    while True:
        value = input(prompt)
        if validator(value):
            return value
        if on_validationerror:
            print(on_validationerror)



def get_user_folder():
    if is_windows():
        return Path(os.environ['USERPROFILE'])
    else:
        NotImplementedError(f'venvlink v. {__version__} supports only Windows')


def is_windows():
    return sys.platform.startswith('win')





import setuptools
import re


def find_version():
    verstrline = open("venvlink/__version__.py", "rt").read()
    try:
        return re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", verstrline, re.M).group(
            1
        )
    except:
        raise RuntimeError("Could not find version string")


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="venvlink",
    description="""Creating the virtual environment outside of the project folder, still retaining the feel of good ol' "python -m venv venv".""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    version=find_version(),
    author="Niko Pasanen",
    author_email="niko@pasanen.me",
    url="https://github.com/np-8/venvlink",
    packages=setuptools.find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Utilities",
    ],
)